#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <iterator>
#include <sstream>
#include <vector>
#include <algorithm>

using namespace std;

// Custom MPI operation function to sum even elements
void custom_sum_even(void *in, void *inout, int *len, MPI_Datatype *dtype) {
    int *input = (int *) in;
    int *inout_array = (int *) inout;
    int array_len = *len;
    int current_value;
    for (int i = 0; i < array_len; i++) {
        current_value = inout_array[i];
        inout_array[i] = 0;
        if ((input[i] % 2) == 0) {
            inout_array[i] += input[i];
        }
        if ((current_value % 2) == 0) {
            inout_array[i] += current_value;
        }
    }
}

int main(int argc, char **argv) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int array_size = 10;
    int *local_array = (int *) malloc(array_size * sizeof(int));
    int *global_array = (int *) malloc(array_size * sizeof(int));

    vector<int> local_vector;
    // Initialize the local array with some data
    for (int i = 0; i < array_size; i++) {
        local_array[i] = i * (rank + 1) + 1;
        global_array[i] = 0;
        local_vector.push_back(i * (rank + 1) + 1);
    }
    std::stringstream result;
    std::copy(local_vector.begin(), local_vector.end(), std::ostream_iterator<int>(result, " "));
    std::cout << "Rank " << rank << " has array {" << result.str().c_str() << "}" << std::endl;

    // Use MPI_Op to create a custom reduction operation
    MPI_Op custom_op;
    MPI_Op_create(custom_sum_even, 0, &custom_op);

    // Perform the reduction to sum even elements
    MPI_Reduce(local_array, global_array, array_size, MPI_INT, custom_op, 0, MPI_COMM_WORLD);

    MPI_Op_free(&custom_op);

    if (rank == 0) {
        printf("Sum of even elements in the global array: ");
        for (int i = 0; i < array_size; i++) {
            printf("%d ", global_array[i]);
        }
        printf("\n");
    }

    free(local_array);
    free(global_array);

    MPI_Finalize();

    return 0;
}
