#include "mpi.h"
#include <iostream>
#include <vector>
#include <cstdlib>
#include <ctime>
#include <bits/stdc++.h>

int main(int argc, char **argv) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    if (size < 2) {
        std::cerr << "This example requires at least 2 processes." << std::endl;
        MPI_Abort(MPI_COMM_WORLD, 1);
    }

    std::srand(static_cast<unsigned>(std::time(0))); // Initialize random number generator

    // Process 0 (master) receives results from other processes
    if (rank == 0) {
        int num_workers = size - 1;
        std::vector<char> results;

        MPI_Status status;
        for (int i = 0; i < num_workers; i++) {
            // Asynchronously probe for incoming message
            MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

            // Get the source (sender) rank
            int source = status.MPI_SOURCE;

            // Get the size of the message (data length)
            int data_length;
            MPI_Get_count(&status, MPI_CHAR, &data_length);

            // Allocate memory for the message
            char *buffer = new char[data_length];

            // Receive the message from the source process
            MPI_Recv(buffer, data_length, MPI_CHAR, source, MPI_ANY_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

            // Concatenate the received message to the results
            results.insert(results.end(), buffer, buffer + data_length);

            std::cout << "Received result from Process " << source << " " << buffer << " (" << data_length << " bytes)"
                      << std::endl;

            delete[] buffer;
        }

        // Process results or perform further computations here
        std::cout << "Total result length: " << results.size() << " bytes" << std::endl;
    }
        // Other processes (workers) perform some computations and send results to the master
    else {
        // Simulate some computation and generate a random data length between 10 and 200
        srand(17 * rank);
        int data_length = std::rand() % 19 + 10;
        char *result = new char[data_length];
        for (int i = 0; i < data_length; i++) {
            result[i] = static_cast<char>('A' + (std::rand() % 26));
        }
        std::cout << "Process " << rank << " sending " << result << " " << " (" << data_length << " bytes)"
                  << std::endl;

        // Send the result to the master
        MPI_Send(result, data_length, MPI_CHAR, 0, rank, MPI_COMM_WORLD);
    }

    MPI_Finalize();
    return 0;
}
